/*
 *  mod_wms.c -- Apache sample wms module
 */

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "ap_config.h"
#include "apr.h"
#include "apr_strings.h"

#include "mapserver.h"	/* for mapObj */
#include "maptemplate.h"
#include "cgiutil.h"

module wms_module;

/* This is our main configuration structure. Well, currently we have only
 * one thing to store... the mapObj.
 */
typedef struct {
  mapObj *map;
} wms_dir_config;

/* These are the IO redirection hooks. They are mostly copied over from
 * mapio.c. Note that cbData contains Apache's request_rec!
 */
static int msIO_apacheWrite( void *cbData, void *data, int byteCount )
{
  /* simply use the block writing function which is very similiar to fwrite */
  return ap_rwrite( data, byteCount, (request_rec *)cbData);
}

static int msIO_apacheError( void *cbData, void *data, int byteCount )
{
  /* error reporting is done through the log file... */
  ap_log_error( APLOG_MARK, APLOG_ERR, 0, NULL, (char*)data );
  return strlen( (char*)data );
}

int msIO_installApacheRedirect( request_rec *r )
{
  msIOContext stdout_ctx, stderr_ctx;

  stdout_ctx.label = "apache";
  stdout_ctx.write_channel = MS_TRUE;
  stdout_ctx.readWriteFunc = msIO_apacheWrite;
  stdout_ctx.cbData = (void *) r;

  stderr_ctx.label = "apache";
  stderr_ctx.write_channel = MS_TRUE;
  stderr_ctx.readWriteFunc = msIO_apacheError;
  stderr_ctx.cbData = (void *) r;

  msIO_installHandlers( NULL, &stdout_ctx, &stderr_ctx );

  return MS_TRUE;
}

/* The maximum number of arguments we support. The number is taken from
 * cgiutil.h I think. It is only used for easier param decoding.
 */
#define WMS_MAX_ARGS 100

/* This is our simple query decoding function. Should be improved but for now
 * it works...
 */
static int
wms_decode_args( apr_pool_t *p, char *args,
                 char ***ParamNames, char ***ParamValues )
{
  char **argv = NULL;
  int i, n;
  int argc = 0;
  char *sep;

  /* No arguments? Then we're done */
  if ( !args ) return 0;

  /* alloc the name/value pointer list */
  argv = apr_pcalloc( p, ( WMS_MAX_ARGS + 1 ) * 2 * sizeof( char* ));
  argv[ 0 ] = args;

  /* separate the arguments */
  for( i = 1, n = 0; args[ n ] && ( i < WMS_MAX_ARGS ); n++ )
    if( args[ n ] == '&' ) {
      argv[ i++ ] = args + n + 1;
      args[ n ] = '\0';
    }

  /* eleminate empty args */
  for( n = 0, i = 0; argv[i]; i++ )
    if( *(argv[i]) != '\0' )
      argv[ n++ ] = argv[ i ];
    else
      argv[ i ] = NULL;
  /* argument count is the number of non-zero arguments! */
  argc = n;

  /* split the name/value pairs */
  for( i = 0; argv[ i ]; i++ ) {
    sep = strchr( argv[ i ], '=' );
    if( !sep ) continue;
    *sep = '\0';
    argv[ i + WMS_MAX_ARGS + 1 ] = (char*)apr_pstrdup( p, sep + 1 );

    if( ap_unescape_url( argv[ i + WMS_MAX_ARGS + 1 ] ) == HTTP_BAD_REQUEST ) {
      ap_log_error( APLOG_MARK, APLOG_ERR, 0, NULL,
                    "%s: malformed URI, couldn't unescape parm %s",
                    __func__, argv[i] );

      argv[ i + WMS_MAX_ARGS + 1 ] = NULL;
    }
  }
  *ParamNames  = argv;
  *ParamValues = argv + WMS_MAX_ARGS + 1;
  return argc;
}

/* That's the main request handler. It's argument is a structure holding
 * everything belonging to the request...
 */
static int
wms_handler( request_rec *r )
{
  /* If the request isn't for us just return */
  if( strcmp( r->handler, "wms" ) != 0 ) return DECLINED;

  /* aquire the apropriate configuration for this directory */
  wms_dir_config *conf = ap_get_module_config( r->per_dir_config,
                                               &wms_module );
  /* if there's no map, something definitely went wrong here */
  if( !conf->map ) {
    ap_log_error( APLOG_MARK, APLOG_ERR, 0, NULL, "No map-object from config" );
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  int argc;
  char **ParamNames;
  char **ParamValues;
  /* Try decoding the query string */
  argc = wms_decode_args( r->pool, (char*)apr_pstrdup( r->pool, r->args ),
                          &ParamNames, &ParamValues );

  if( !argc ) {
    return HTTP_BAD_REQUEST;
  }

  int i, j, k, status;
  /* we don't really need the msObj itself I think but I wanted to modify
   * the code as little as possible to show the similarity */
  mapservObj *msObj = msAllocMapServObj();
  /* we fill in the those values which were normally inserted by the CGI
   * functions...
   */
  msObj->request->NumParams = argc;
  msObj->request->ParamNames = ParamNames;
  msObj->request->ParamValues = ParamValues;
  /* OK, here's the magic: we take the mapObj from our stored config.
   * It should be noted that we really should make a copy of it since many
   * functions will modify the mapObj - for setting to output type for
   * example. With the current situation we can _not_ use the MPM WORKER model
   * of Apache since this would lead to race conditions. Therefore you have
   * to use the MPM PREFORK model for now...
   */
  msObj->Map = conf->map;

  char *key, *value, *tmpstr;
  /* Now we install the IO redirection. Not sure if it must be done on a
   * per request basis.
   */
  if( msIO_installApacheRedirect( r ) != MS_TRUE ) {
    ap_log_error( APLOG_MARK, APLOG_ERR, 0, NULL, "%s: could not install apache redirect", __func__ );
  }
  /* Look carefully! It really looks like mapserv.c code, doesn't it?
   * It's simply copied over. The TEST_FOR macro was put in for better
   * readability only. The first part dealing with the map parameter was
   * omitted. Now follows the part of the LoadMap() function...
   */
#define TEST_FOR( name, len ) ( strncasecmp( ParamNames[ i ], name, len ) == 0 )

  for( i = 0; i < argc; i++ ) {
    if( TEST_FOR( "map_", 4 ) || TEST_FOR( "map.", 4 )) {
      if( msUpdateMapFromURL( msObj->Map, ParamNames[ i ],
                              ParamValues[ i ]) != MS_SUCCESS ) {
        continue;
      }
    }
    tmpstr = (char*)apr_psprintf( r->pool, "%%%s%%", ParamNames[ i ]);
    key = (char*)apr_psprintf( r->pool, "%s_validation_pattern", ParamNames[ i ]);
    for( j = 0; j < msObj->Map->numlayers; j++ ) {
      value = msLookupHashTable(&( GET_LAYER( msObj->Map, j )->metadata ), key );
      if( value ) /* validate parameter value */
        if( msEvalRegex( value, ParamValues[ i ]) == MS_FALSE ) {
          ap_log_error( APLOG_MARK, APLOG_ERR, 0, NULL, "Malformed URI" );
        }

      if( GET_LAYER( msObj->Map, j )->data &&
          strstr( GET_LAYER( msObj->Map, j )->data, tmpstr ))
        GET_LAYER( msObj->Map, j )->data =
          msReplaceSubstring( GET_LAYER( msObj->Map, j )->data, tmpstr,
                              ParamValues[ i ]);
      if( GET_LAYER( msObj->Map, j )->tileindex &&
          strstr( GET_LAYER( msObj->Map, j )->tileindex, tmpstr ))
        GET_LAYER( msObj->Map, j )->tileindex =
          msReplaceSubstring( GET_LAYER( msObj->Map, j )->tileindex, tmpstr,
                              ParamValues[ i ]);
      if( GET_LAYER( msObj->Map, j )->connection &&
          strstr( GET_LAYER( msObj->Map, j )->connection, tmpstr ))
        GET_LAYER( msObj->Map, j )->connection =
          msReplaceSubstring( GET_LAYER( msObj->Map, j )->connection, tmpstr,
                              ParamValues[ i ]);
      if( GET_LAYER( msObj->Map, j )->filter.string &&
          strstr( GET_LAYER( msObj->Map, j )->filter.string, tmpstr ))
        GET_LAYER( msObj->Map, j )->filter.string =
          msReplaceSubstring( GET_LAYER( msObj->Map, j )->filter.string, tmpstr,
                              ParamValues[ i ]);
      for( k = 0; k < GET_LAYER( msObj->Map, j )->numclasses; k++ )
        if( GET_LAYER( msObj->Map, j )->class[ k ]->expression.string &&
            strstr( GET_LAYER( msObj->Map, j )->class[ k ]->expression.string,
                     tmpstr ))
          GET_LAYER( msObj->Map, j )->class[ k ]->expression.string =
            msReplaceSubstring( GET_LAYER( msObj->Map, j )->class[ k ]->
                                expression.string, tmpstr, ParamValues[ i ]);
    }
  }
  for( i = 0; i < argc; i++ )
    if( !strcasecmp( ParamNames[ i ], "context" ))
      if( ParamValues[ i ] && strlen( ParamValues[ i ]) > 0 ) {
        if( !strncasecmp( ParamValues[i], "http", 4 )) {
          if( msGetConfigOption( msObj->Map, "CGI_CONTEXT_URL" ))
            msLoadMapContextURL( msObj->Map, ParamValues[ i ], MS_FALSE );
        } else
            msLoadMapContext( msObj->Map, ParamValues[ i ], MS_FALSE );
      }

  /* Now we're right back from LoadMap() and go on with the main mapserv code.
   * Well, there's nothing more to do than jumping into the OWS dispatcher.
   */
  if(( status = msOWSDispatch( msObj->Map, msObj->request )) != MS_DONE  ) {
    /*
     * Normally if the OWS service fails it will issue an exception,
     * and clear the error stack but still return MS_FAILURE.  But in
     * a few situations it can't issue the exception and will instead
     * just an error and let us report it.
     */
    if( status == MS_FAILURE ) {
      errorObj *ms_error = msGetErrorObj();

      if( ms_error->code != MS_NOERR )
        ap_log_error( APLOG_MARK, APLOG_ERR, 0, NULL,
                      "Error in msOWSDispatch" );
    }

    /* normal case, processing is complete */
    /* OK, here we are again. Since the CGI related things were done by us
     * we also must take care that the cleanup functions don't touch that
     * stuff. Apache will free the memory when destroying the pool as a whole.
     * It should be noted that this is _not_ a bad programming style! Instead
     * Apache uses the memory pooling to prevent fragmentation. We also
     * want to preserve our mapObj for the next request...
     */
    msObj->Map = NULL;
    msObj->request->ParamNames = NULL;
    msObj->request->ParamValues = NULL;
    msObj->request->NumParams = 0;
    msFreeMapServObj(msObj);
    return OK;
  } /* done OGC/OWS case */

  /* Now follows the useless part of the module template which is reached if
   * it's not a OWS request. We can return DECLINED to tell Apache he should
   * try another handler and if there's none we finally reach HTTP 404 - not
   * found. Or we could return HTTP_BAD_REQUEST to show that something was
   * wrong. It's up to you and may be configurable...
   */
  r->content_type = "text/html";

  if( r->header_only ) return OK;
    ap_rputs( "The sample page from mod_wms.c\n", r );
  return OK;
}

/* This function will be called on startup to allow us to register our
 * handler. We don't do anything else yet...
 */
static void
wms_register_hooks( apr_pool_t *p )
{
  ap_hook_handler( wms_handler, NULL, NULL, APR_HOOK_MIDDLE );
}

/* That's the second part of the overall magic. This function will be called
 * when the configuration reaches the "WMS_Map" parameter. It will be passed
 * in as argument and should contain the full path and name of the map file.
 * We try to load it and store the resulting mapObj in the config of the
 * current directory. If we fail we throw a message...
 */
static const char*
wms_set_map( cmd_parms *cmd, void *config, const char *arg )
{
  wms_dir_config *conf = (wms_dir_config*)config;
  /* if the mapObj already exists the WMS_Map was given more than once -
   * may be the user forgot to comment something out...
   */
  if( conf->map )
    return (char*)apr_psprintf( cmd->temp_pool,
                                "An MAP-file has already been registered for "
                                "this URI - not accepting '%s'.", arg );
  /* Simply try loading the argument as map file. */
  conf->map = msLoadMap( (char*)arg, NULL );
  /* Ooops - we failed. We report it... */
  if( !conf->map )
    return (char*)apr_psprintf( cmd->temp_pool,
                                "The given MAP-file '%s' could not be loaded",
                                arg );
  return NULL;
}

/* This function creates the (default) directory configuration. Well, we have
 * no defaults to set so we simply alloc the memory...
 */
static void*
wms_create_dir_config( apr_pool_t *p, char *dir )
{
  wms_dir_config *newconf;
  newconf = (wms_dir_config*) apr_pcalloc( p, sizeof( wms_dir_config ));
  return newconf;
}

/* This structure defines our single config option which takes exactly one
 * argument.
 */
static const command_rec
wms_options[] = {
  AP_INIT_TAKE1(
    "WMS_Map",
    wms_set_map,
    NULL,
    ACCESS_CONF,
    "WMS_Map <Path to map file>"
  ),
  {NULL}
};

/* The following structure declares everything Apache needs to treat us as
 * a module. Merging would mean inheriting from the parent directory but that
 * doesn't make much sense here since we would only inherit the map file...
 */
/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA
wms_module = {
    STANDARD20_MODULE_STUFF,
    wms_create_dir_config,	/* create per-dir    config structures */
    NULL,			/* merge  per-dir    config structures */
    NULL,			/* create per-server config structures */
    NULL,			/* merge  per-server config structures */
    wms_options,		/* table of config file commands       */
    wms_register_hooks		/* register hooks                      */
};
